from contextlib import contextmanager
from time import perf_counter


def pretty_print_time(seconds):
    # Calculate minutes and seconds
    minutes, secs = divmod(seconds, 60)
    # Format the time as mm:ss
    return f"{int(minutes):02d}:{int(secs):02d}"


@contextmanager
def timing(name="Block"):
    start = perf_counter()
    yield
    end = perf_counter()
    print(f"[{name}] took {end - start:.6f} seconds")



import logging
import sys
from contextlib import contextmanager
from logging import Logger
from pathlib import Path
from typing import Optional


class ColorFormatter(logging.Formatter):
    RESET = "\033[0m"
    COLORS = {
        logging.DEBUG: "\033[34m",  # Blue
        logging.WARNING: "\033[31m",  # Red
    }

    def format(self, record):
        color = self.COLORS.get(record.levelno, "")
        message = super().format(record)
        return f"{color}{message}{self.RESET}" if color else message


class StdoutFilter(logging.Filter):
    def filter(self, record):
        return record.levelno < logging.ERROR  # DEBUG, INFO, WARNING


class StderrFilter(logging.Filter):
    def filter(self, record):
        return record.levelno >= logging.ERROR  # ERROR, CRITICAL


def get_stdout_handler(formatter):
    # Stdout handler (DEBUG, INFO, WARNING)
    stdout_handler = logging.StreamHandler(sys.stdout)
    stdout_handler.setLevel(logging.DEBUG)
    stdout_handler.addFilter(StdoutFilter())
    stdout_handler.setFormatter(formatter)
    return stdout_handler


def get_stderr_handler(formatter):
    # Stderr handler (ERROR, CRITICAL)
    stderr_handler = logging.StreamHandler(sys.stderr)
    stderr_handler.setLevel(logging.ERROR)
    stderr_handler.addFilter(StderrFilter())
    stderr_handler.setFormatter(formatter)
    return stderr_handler


class logger:  # noqa
    client: Optional[Logger] = None
    section = 'main'
    ENABLED_SECTIONS = {
        'main': True,
        'game': True,
        'monte_carlo': True,
        'alpha_beta': True,
        'training': True,
        'benchmark': True,
    }
    history = []

    @classmethod
    def setup(
        cls,
        name: str = None,
        loglevel: str = 'INFO',
        section: str | list = None,
    ):
        """
        Sets up the logger.

        :param name: Name of the logger
        :param loglevel: log level (default: INFO)
        :param section: sections of the logger to render (default: all)
        """
        cls.set_section(section)

        if cls.client is not None and cls.client.level == logging.INFO:
            return
        cls.client = logging.getLogger(name)
        cls.client.setLevel(loglevel)

        formatter = ColorFormatter(
            # '%(asctime)s - '
            # '%(name)s - '
            # '%(levelname)s - '
            '%(message)s'
        )

        stdout_handler = get_stdout_handler(formatter)
        stderr_handler = get_stderr_handler(formatter)

        cls.client.handlers = []
        cls.client.addHandler(stdout_handler)
        cls.client.addHandler(stderr_handler)

    @classmethod
    @contextmanager
    def setup_in_context(cls, *args, **kwargs):
        cls.setup(*args, **kwargs)
        enabled_sections = logger.ENABLED_SECTIONS
        try:
            yield
        finally:
            cls.stop()
            logger.ENABLED_SECTIONS = enabled_sections
            cls.clear_history()

    @classmethod
    def stop(cls):
        logger.client = None

    @classmethod
    def set_section(cls, section: str | list) -> None:
        """
        Sets the sections of the logger.

        :param section: Section of the logger to render (default: all)
        """
        if not section:
            return
        if isinstance(section, str):
            section = [section]
        for s in section:
            assert s in cls.ENABLED_SECTIONS, f"Section {s} not in {list(cls.ENABLED_SECTIONS.keys())}"
        for k, v in cls.ENABLED_SECTIONS.items():
            cls.ENABLED_SECTIONS[k] = k in section

    @classmethod
    def clear_history(cls) -> None:
        logger.history = []

    @classmethod
    def dump_history(cls, path: Path | str = None, mode='a', clear=False) -> None:
        """
        Dump log history to a text file.

        :param path: Path to the text file (default: None)
        :param mode: How to handle if the file exists (default: 'a')
        :param clear: Whether to clear the history after dumping (default: False)
        """
        if cls.client is None:
            return
        text = '\n'.join(logger.history) + '\n'
        with open(path, mode=mode) as f:
            f.write(text)
        if clear:
            cls.clear_history()

    @classmethod
    def log(cls, msg, level=logging.INFO, **kwargs):
        if (
            cls.client is not None
            and (cls.ENABLED_SECTIONS[cls.section] or level > logging.INFO)
            and cls.client.isEnabledFor(level)
        ):
            cls.client.log(msg=msg, level=level, stacklevel=3, **kwargs)
            logger.history.append(str(msg))

    @classmethod
    def debug(cls, msg, **kwargs):
        cls.log(msg=msg, level=logging.DEBUG, **kwargs)

    @classmethod
    def info(cls, msg, **kwargs):
        cls.log(msg=msg, **kwargs)

    @classmethod
    def warn(cls, msg, **kwargs):
        cls.log(msg=msg, level=logging.WARNING, **kwargs)

    @classmethod
    def error(cls, msg, **kwargs):
        cls.log(msg=msg, level=logging.ERROR, **kwargs)

    @classmethod
    def critical(cls, msg, **kwargs):
        cls.log(msg=msg, level=logging.CRITICAL, **kwargs)

    @classmethod
    @contextmanager
    def context_info(cls, msg: str):
        """
        Log start and end of the block
        """
        cls.info(msg)
        yield
        cls.info(f'... {msg} done')


class game_logger(logger):  # noqa
    section = 'game'


class monte_carlo_logger(logger):  # noqa
    section = 'monte_carlo'


class alpha_beta_logger(logger):  # noqa
    section = 'alpha_beta'


class training_logger(logger):  # noqa
    section = 'training'


class benchmark_logger(logger):  # noqa
    section = 'benchmark'



import json
import pickle
from pathlib import Path


def extend_filename(file_path: str | Path, s: str) -> Path:
    """
    Extends the filename of a file with a string.
    :param file_path:
    :param s:
    :return:

    Example:
    >>> Path(extend_filename("path/to/file.txt", "_new")).as_posix()
    'path/to/file_new.txt'

    >>> extend_filename(Path("file.txt"), "_old").as_posix()
    'file_old.txt'
    """
    is_string = isinstance(file_path, str)
    file_path = Path(file_path)
    file_path_new = file_path.with_name(file_path.stem + s + file_path.suffix)
    if is_string:
        file_path_new = str(file_path_new)
    return file_path_new


def dump_pickle(obj, file_path: str | Path):
    """
    Dumps an object to a file using pickle.
    :param obj:
    :param file_path:
    :return:
    """

    with open(file_path, 'wb') as f:
        pickle.dump(obj, f)  # noqa


def load_pickle(file_path: str | Path, raise_error=True):
    """
    Loads an object from a file using pickle.
    :param file_path:
    :param raise_error: If True, raises an error if the file doesn't exist. If False, returns None.
    :return:
    """
    if not Path(file_path).exists():
        if raise_error:
            raise FileNotFoundError(f"File {file_path} not found.")
        return None

    try:
        with open(file_path, 'rb') as f:
            return pickle.load(f)
    except (EOFError, pickle.UnpicklingError):
        if raise_error:
            raise
        return None


def dump_json(obj, file_path: str | Path):
    """
    Dumps an object to a file using pickle.
    :param obj:
    :param file_path:
    :return:
    """
    with open(file_path, 'w') as f:
        json.dump(obj, f)


def load_json(file_path: str | Path, raise_error=True):
    """
    Loads an object from a file using pickle.
    :param file_path:
    :param raise_error: If True, raises an error if the file doesn't exist. If False, returns None.
    :return:
    """
    if not Path(file_path).exists():
        if raise_error:
            raise FileNotFoundError(f"File {file_path} not found.")
        return None

    try:
        with open(file_path, 'r') as f:
            return json.load(f)
    except json.JSONDecodeError:
        if raise_error:
            raise
        return None


def load_txt(file_path: str | Path):
    """
    Loads an object from a file using pickle.
    :param file_path:
    :return:
    """
    with open(file_path, 'r') as f:
        return f.read()


def dump_txt(obj, file_path: str | Path, mode='w'):
    """
    Dumps an object to a file.
    :param obj:
    :param file_path:
    """
    with open(file_path, mode=mode) as f:
        f.write(obj)


def mkdir(path: str | Path):
    """
    Creates a directory if it doesn't exist.
    :param path:
    """
    Path(path).mkdir(parents=True, exist_ok=True)


def human_readable_size(size_in_bytes):
    for unit in ['B', 'KB', 'MB', 'GB']:
        if size_in_bytes < 1024:
            return f"{size_in_bytes:.2f} {unit}"
        size_in_bytes /= 1024
    return f"{size_in_bytes:.2f} TB"



from collections import defaultdict

import numpy as np

class PrettyDict(dict):
    def __repr__(self):
        return ', '.join([f"{k}={v:.1f}" for k, v in self.items()])


def dict_factory():
    return defaultdict(dict_factory)


def check_nan(data):
    """
    >>> check_nan(None)
    >>> check_nan(np.array([1, 2, 3]))
    >>> check_nan(np.array([1, float('nan'), 3]))
    Traceback (most recent call last):
       ...
    RuntimeError: NaN detected in: [ 1. nan  3.]
    """
    if data is None:
        return
    if isinstance(data, np.ndarray) and not np.isnan(data).any():
        return
    raise RuntimeError(f"NaN detected in: {data}")



import platform
import subprocess

import psutil


def get_cpu_info():
    # CPU brand and speed (cross-platform)
    cpu = platform.processor() or "Unknown CPU"
    try:
        import cpuinfo
        cpu = cpuinfo.get_cpu_info()['brand_raw']
    except ImportError:
        print("Note: You can 'pip install py-cpuinfo' for better CPU info")
        pass

    # Core/thread info
    cores = psutil.cpu_count(logical=False) or 0
    threads = psutil.cpu_count(logical=True) or 0

    return f"{cpu} ({cores} cores, {threads} threads)"


def get_ram_info():
    mem = psutil.virtual_memory()
    total_gb = round(mem.total / (1024 ** 3), 2)
    return f"{total_gb} GB RAM"


def get_gpu_info():
    # Try NVIDIA first (requires nvidia-smi)
    try:
        result = subprocess.run(
            ["nvidia-smi", "--query-gpu=name,memory.total,driver_version", "--format=csv,noheader"],
            stdout=subprocess.PIPE, stderr=subprocess.DEVNULL, text=True, check=True
        )
        gpus = []
        for line in result.stdout.strip().splitlines():
            name, mem, driver = [x.strip() for x in line.split(",")]
            gpus.append(f"{name} ({mem} VRAM, Driver {driver})")
        return "; ".join(gpus)
    except Exception:
        return "No NVIDIA GPU detected"


def get_os_info():
    return f"{platform.system()} {platform.release()} ({platform.version()})"


def get_python_info():
    return f"Python {platform.python_version()} ({platform.python_implementation()})"


def get_library_versions():
    libs = {}
    for lib in ["numpy", "torch", "tensorflow"]:
        try:
            module = __import__(lib)
            libs[lib] = module.__version__
        except ImportError:
            libs[lib] = "Not installed"
    return libs


def get_system_report():
    report = {
        "CPU": get_cpu_info(),
        "GPU": get_gpu_info(),
        "RAM": get_ram_info(),
        "OS": get_os_info(),
        "Python": get_python_info(),
        "Libraries": get_library_versions(),
    }
    return report



import sys
from contextlib import contextmanager
from dataclasses import dataclass
from pathlib import Path

import gcaa

EPS = 1e-12


@dataclass
class DefaultParams:
    time_out = 3600.0

    def __str__(self):
        return str(self.__dict__)

    @classmethod
    def attributes(cls):
        return {
            k: getattr(cls, k)
            for k in dir(cls)
            if not k.startswith('_') and not callable(getattr(cls, k))
        }

    @classmethod
    def print(cls):
        print(cls.attributes())

    @classmethod
    @contextmanager
    def update(cls, **kwargs):
        old = cls.attributes()
        for key, value in kwargs.items():
            setattr(cls, key, value)
        try:
            yield
        finally:
            for key, value in old.items():
                setattr(cls, key, value)


RESOURCE_PATH = Path(gcaa.__file__).parent / 'resources'
DATA_PATH = Path(gcaa.__file__).parent / 'data'
MAX_INT = sys.maxsize
inf = float("inf")



import os
import time
from datetime import datetime

FILE_DATE_FMT = "%Y%m%d_%H%M%S"
READABLE_DATE_FMT = '%Y-%m-%d %H:%M:%S'


def get_now(human_readable=True, fmt=FILE_DATE_FMT):
    now = datetime.now()
    if human_readable:
        now = now.strftime(fmt)
    return now


def get_file_modified_time(filepath, fmt=READABLE_DATE_FMT):
    return time.strftime(
        fmt,
        time.localtime(os.path.getmtime(filepath))
    )



from IPython import get_ipython


def is_notebook() -> bool:
    """
    >>> is_notebook()
    False
    """
    try:
        shell = get_ipython().__class__.__name__
        if shell == 'ZMQInteractiveShell':
            return True  # Jupyter notebook or qtconsole
        elif shell == 'TerminalInteractiveShell':
            return False  # Terminal running IPython
        else:
            return False  # Another type
    except NameError:
        return False  # Probably standard Python interpreter



def hash_dict(d):
    """Generates a hash for a dictionary, including nested dictionaries."""
    # May want to use `hashlib.sha256(s).hexdigest()` in the future for reproducibility and
    # lower risk of collision
    # If the input is a dictionary, convert it to a sorted tuple of key-hash pairs
    if isinstance(d, dict):
        return hash(tuple(sorted((key, hash_dict(value)) for key, value in d.items())))
    # If the input is a list or tuple, hash its immutable version
    elif isinstance(d, (list, tuple)):
        return hash(tuple(hash_dict(item) for item in d))
    # For anything else, return its direct hash
    else:
        return hash(d)



import numpy as np


def box(x):
    if isinstance(x, list):
        return x
    return [x]


def unbox(x):
    if isinstance(x, list) and len(x) == 1:
        x = x[0]
    return x


def array2string(arr):
    formatter = {'float_kind': lambda x: f"{x:.2f}"}
    return np.array2string(
        arr,
        formatter=formatter  # noqa
    )
